#ifndef STAN_LANG_AST_NODE_NO_OP_STATEMENT_HPP
#define STAN_LANG_AST_NODE_NO_OP_STATEMENT_HPP

namespace stan {
  namespace lang {

    /**
     * AST node for the no-operation statement.
     */
    struct no_op_statement { };

  }
}
#endif
