// Copyright Antony Polukhin, 2016-2018.
//
// Distributed under the Boost Software License, Version 1.0. (See
// accompanying file LICENSE_1_0.txt or copy at
// http://www.boost.org/LICENSE_1_0.txt)

// No include guards! Intentionally.

#ifdef BOOST_STACKTRACE_FUNCTION
#   undef BOOST_STACKTRACE_FUNCTION
#endif

