//
// Copyright (c) 2016-2017 Vinnie Falco (vinnie dot falco at gmail dot com)
//
// Distributed under the Boost Software License, Version 1.0. (See accompanying
// file LICENSE_1_0.txt or copy at http://www.boost.org/LICENSE_1_0.txt)
//
// Official repository: https://github.com/boostorg/beast
//

#ifndef BOOST_BEAST_ZLIB_HPP
#define BOOST_BEAST_ZLIB_HPP

#include <boost/beast/core/detail/config.hpp>

#include <boost/beast/zlib/deflate_stream.hpp>
#include <boost/beast/zlib/error.hpp>
#include <boost/beast/zlib/inflate_stream.hpp>
#include <boost/beast/zlib/zlib.hpp>

#endif
