#include <stan/lang/grammars/indexes_grammar_def.hpp>
#include <stan/lang/grammars/iterator_typedefs.hpp>

namespace stan {
  namespace lang {

    template struct stan::lang::indexes_grammar<pos_iterator_t>;

  }
}
